#include "wool/wool.h"
#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>

static __inline__ uint64_t getticks(void);
void printErrorIfArrayIsNotSorted(int * array, int size);
char isSorted(int * array, int size);
void merge(int * firstPart, int * secondPart, int * mergedArray, int firstSize, int secondSize);
void fillWithRandomValues(int * array_to_fill, int size);
int compare(const void * a, const void * b);
void swap(int ** a, int ** b);
void printArray(int * array, int size);
int binarySearch(int * array, int first, int last, int key);

#ifndef min
#define min(a, b) ( ((a) < (b)) ? (a) : (b) )
#endif

int problem_size = 10000000;
int cutOff = 75000;
int mergeCutOff = 75000;

TASK_2( int, main, int, argc, char **, argv )
{

	if (argc >= 2) {
		problem_size = atoi(argv[1]);
	}

	if (argc >= 3) {
		cutOff = atoi(argv[2]);
	}

	if (argc >= 4) {
		mergeCutOff = atoi(argv[3]);
	}

	int * scrambled = malloc(problem_size * sizeof(int));
	int * sorted = malloc(problem_size * sizeof(int));

	uint64_t pPre, pPost;

	fillWithRandomValues(scrambled, problem_size);

	pPre = getticks();
	CALL( psort, &scrambled, &sorted, problem_size);
	pPost = getticks();

	uint64_t pCycles = pPost - pPre;

	printf("pCycles=%"PRIu64"\n", pCycles);

	printErrorIfArrayIsNotSorted(sorted, problem_size);

	free(scrambled);
	free(sorted);

	return 0;
}

VOID_TASK_3(psort, int**, in, int**, out, int, size )
{
	if (size <= cutOff) {
		qsort(*in, size, sizeof(int), compare);
		swap(in, out);
	} else {

		int firstProblemSize = size / 2;
		int secondProblemSize = size - firstProblemSize;

		int * secondInput = &((*in)[firstProblemSize]);
		int * secondOutput = &((*out)[firstProblemSize]);

		SPAWN( psort, in, out, firstProblemSize );
		CALL( psort, &secondInput, &secondOutput, secondProblemSize );
		SYNC( psort );

		swap(in, out);

		//This is useful when seeing if the profiler can pick up on the badly optimized merge sort.
		merge(*in, &(*in)[firstProblemSize], *out, firstProblemSize, secondProblemSize);

		//CALL( pmerge, *in, &(*in)[firstProblemSize], *out, firstProblemSize, secondProblemSize );
	}
}

VOID_TASK_5( pmerge , int*, firstPart, int*, secondPart, int*, mergedArray, int, firstSize, int, secondSize)
{
	if (firstSize + secondSize <= mergeCutOff) {
		merge(firstPart, secondPart, mergedArray, firstSize, secondSize);
	} else {

		if (firstSize < secondSize) { //swap them if the second is bigger

			int tempSize = firstSize;
			firstSize = secondSize;
			secondSize = tempSize;

			int * tempPart = firstPart;
			firstPart = secondPart;
			secondPart = tempPart;
		}
		int midOfFirst = firstSize / 2;

		int cutOffInSecond = min(binarySearch(secondPart, 0, secondSize, firstPart[midOfFirst]), secondSize);

		if (secondSize == 0) {
			//just return first

			for (int i = 0; i < firstSize; ++i) {
				mergedArray[i] = firstPart[i];
			}
			return;
		}

		int * tempPointer = &(mergedArray[midOfFirst + 1 + cutOffInSecond]);

		SPAWN( pmerge, firstPart, secondPart, mergedArray, midOfFirst+1, cutOffInSecond );
		CALL( pmerge, &(firstPart[midOfFirst + 1]), &(secondPart[cutOffInSecond]), tempPointer, firstSize - (midOfFirst + 1), secondSize - cutOffInSecond);

		SYNC( pmerge );
	}
	return;
}

static __inline__ uint64_t getticks(void) {

	unsigned int a, d;
	asm volatile("rdtsc" : "=a" (a), "=d" (d));
	return ((uint64_t) a) | (((uint64_t) d) << 32);
}

void merge(int * firstPart, int * secondPart, int * mergedArray, int firstSize, int secondSize) {

	int firstCounter = 0;
	int secondCounter = 0;

	while (firstCounter < firstSize && secondCounter < secondSize) {
		int mergedCounter = firstCounter + secondCounter;

		if (firstPart[firstCounter] < secondPart[secondCounter]) {
			mergedArray[mergedCounter] = firstPart[firstCounter];
			++firstCounter;
		} else {
			mergedArray[mergedCounter] = secondPart[secondCounter];
			++secondCounter;
		}
	}

	while (firstCounter < firstSize) {
		mergedArray[firstCounter + secondCounter] = firstPart[firstCounter];
		++firstCounter;
	}

	while (secondCounter < secondSize) {
		mergedArray[firstCounter + secondCounter] = secondPart[secondCounter];
		++secondCounter;
	}
}

void printErrorIfArrayIsNotSorted(int * array, int size) {
	if (!isSorted(array, size)) {
		fprintf(stderr, "ERROR --- ERROR \nThe array is not sorted \nERROR --- ERROR\n");
	}
}

char isSorted(int * array, int size) {
	for (int i = 0; i < size - 1; i++) {
		if (array[i] > array[i + 1]) {
			return 0;
		}
	}
	return 1;
}

void fillWithRandomValues(int * array_to_fill, int size) {
	for (int i = 0; i < size; i++) {
		array_to_fill[i] = rand();
	}
}

int compare(const void * a, const void * b) {
	return (*(int*) a - *(int*) b);
}

void swap(int ** a, int ** b) {
	int * temp = *a;
	*a = *b;
	*b = temp;
}

int binarySearch(int * array, int first, int last, int key) {
	while (first <= last) {
		int mid = (first + last) / 2;
		if (key > array[mid])
			first = mid + 1;
		else if (key < array[mid])
			last = mid - 1;
		else
			return mid;
	}
	return first;
}
