package gui;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.io.IOException;
import java.util.Hashtable;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSlider;
import javax.swing.UIManager;
import javax.swing.UIManager.LookAndFeelInfo;

import util.StatsReader;

@SuppressWarnings("serial")
public class Main extends JFrame {

	public static void main(String[] args) throws Exception {

		for (LookAndFeelInfo info : UIManager.getInstalledLookAndFeels()) {
			if ("Nimbus".equals(info.getName())) {
				UIManager.setLookAndFeel(info.getClassName());
				break;
			}
		}

		new Main();

	}

	public Main() throws IOException {
		super("WoolPlot");
		setLayout(new GridBagLayout());

		StatsReader sr = new StatsReader("e-steal100");

		final PlotPanel plotPanel = sr.makePlotPanel();

		GridBagConstraints constraints = new GridBagConstraints();

		final JButton stealButton = createStealButton(plotPanel);

		final JButton leapButton = createLeapButton(plotPanel);

		final JButton spawnButton = createSpawnButton(plotPanel);

		final JButton usageButton = createUsageButton(plotPanel);

		final JButton criticalPathButton = createCriticalPathButton(plotPanel);

		JPanel buttonPanel = new JPanel();
		buttonPanel.add(stealButton);
		buttonPanel.add(leapButton);
		buttonPanel.add(spawnButton);
		buttonPanel.add(usageButton);
		buttonPanel.add(criticalPathButton);

		constraints.gridy = 0;
		add(buttonPanel, constraints);

		final JScrollPane jScrollPane = new JScrollPane(plotPanel, JScrollPane.VERTICAL_SCROLLBAR_NEVER, JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);

		jScrollPane.setDoubleBuffered(true);

		jScrollPane.addMouseWheelListener(new MouseWheelListener() {
			@Override
			public void mouseWheelMoved(MouseWheelEvent e) {
				jScrollPane.getHorizontalScrollBar().setValue(
						(jScrollPane.getHorizontalScrollBar().getValue() - e.getWheelRotation() * (plotPanel.getWidth() / 100)));
			}
		});

		constraints.gridy = 1;

		constraints.fill = GridBagConstraints.BOTH;
		constraints.weightx = 1;
		constraints.weighty = 1;
		add(jScrollPane, constraints);

		PanelModel panelModel = new PanelModel(plotPanel.getWidth(), 0, plotPanel.getWidth(), 60000);
		plotPanel.setModel(panelModel);

		final JSlider slider = new JSlider(panelModel);

		setUpSlider(slider);

		constraints.weighty = 0;
		constraints.gridy = 2;
		add(slider, constraints);

		pack();
		setVisible(true);
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		setLocationRelativeTo(null);
	}

	private JButton createStealButton(final PlotPanel plotPanel) {
		final JButton stealButton = new JButton("Hide steals");

		stealButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				plotPanel.toggleStealPainting();
				plotPanel.repaint();
				if (arg0.getActionCommand().startsWith("Hide")) {
					stealButton.setText("Paint steals");
				} else {
					stealButton.setText("Hide steals");
				}
			}
		});
		return stealButton;
	}

	private JButton createLeapButton(final PlotPanel plotPanel) {
		final JButton leapButton = new JButton("Hide leaps");

		leapButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				plotPanel.toggleLeapPainting();
				plotPanel.repaint();
				if (arg0.getActionCommand().startsWith("Hide")) {
					leapButton.setText("Paint leaps");
				} else {
					leapButton.setText("Hide leaps");
				}

			}
		});
		return leapButton;
	}

	private JButton createSpawnButton(final PlotPanel plotPanel) {
		final JButton spawnButton = new JButton("Hide spawns");

		spawnButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				plotPanel.toggleSpawnPainting();
				plotPanel.repaint();
				if (e.getActionCommand().startsWith("Hide")) {
					spawnButton.setText("Paint spawns");
				} else {
					spawnButton.setText("Hide spawns");
				}
			}
		});

		return spawnButton;
	}

	private JButton createUsageButton(final PlotPanel plotPanel) {
		final JButton usageButton = new JButton("Hide usages");

		usageButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				plotPanel.toggleUsagePainting();
				plotPanel.repaint();
				if (e.getActionCommand().startsWith("Hide")) {
					usageButton.setText("Paint usages");
				} else {
					usageButton.setText("Hide usages");
				}
			}
		});

		return usageButton;
	}

	private JButton createCriticalPathButton(final PlotPanel plotPanel) {
		final JButton criticalPathButton = new JButton("Hide critical path");

		criticalPathButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				plotPanel.toggleCriticalPathPainting();
				plotPanel.repaint();
				if (e.getActionCommand().startsWith("Hide")) {
					criticalPathButton.setText("Paint critical path");
				} else {
					criticalPathButton.setText("Hide critical path");
				}
			}
		});

		return criticalPathButton;
	}

	private void setUpSlider(final JSlider slider) {
		setPainting(slider);

		setLabels(slider);

		slider.addMouseWheelListener(new MouseWheelListener() {
			@Override
			public void mouseWheelMoved(MouseWheelEvent e) {
				slider.getModel().setValue(slider.getValue() - e.getWheelRotation() * slider.getMaximum() / 100);
			}
		});
	}

	private void setPainting(final JSlider slider) {
		slider.setMajorTickSpacing((slider.getMaximum() - slider.getMinimum()) / 10);
		slider.setMinorTickSpacing((slider.getMaximum() - slider.getMinimum()) / 100);
		slider.setPaintTicks(true);
		slider.setPaintLabels(true);
	}

	private void setLabels(final JSlider slider) {
		Hashtable<Integer, JLabel> labelTable = new Hashtable<Integer, JLabel>();
		labelTable.put(slider.getMinimum(), new JLabel("0%"));
		labelTable.put(slider.getMinimum() + (slider.getMaximum() - slider.getMinimum()) / 2, new JLabel("50%"));
		labelTable.put(slider.getMaximum(), new JLabel("100%"));
		slider.setLabelTable(labelTable);
	}
}