package util;

import gui.PlotPanel;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collections;

public class StatsReader {

	private static BufferedReader bufferedReader;
	private int numberOfWorkers;
	private BigDecimal startTime;
	private BigDecimal endTime;

	public StatsReader(String fileName) {
		createReader(fileName);
		numberOfWorkers = -1; // While waiting for makePlotCanvas() to set it to the correct value;
	}

	private void createReader(String fileName) {
		FileReader reader = null;

		try {
			reader = new FileReader(fileName);
		} catch (FileNotFoundException e) {
			throw new IllegalArgumentException("File " + fileName + " does not exist.");
		}

		bufferedReader = new BufferedReader(reader);
	}

	public PlotPanel makePlotPanel() throws IOException {

		readCrap();

		numberOfWorkers = Integer.parseInt(bufferedReader.readLine());
		startTime = new BigDecimal(bufferedReader.readLine());
		endTime = new BigDecimal(bufferedReader.readLine());

		ArrayList<ArrayList<Steal>> steals = makeListOfSteals();
		ArrayList<ArrayList<Steal>> leaps = makeListOfSteals();
		ArrayList<ArrayList<Spawn>> spawns = makeListOfSpawns();

		setTheStartTimeToTheFirstSpawn(spawns);

		ArrayList<Event> events = makeListOfEvents();

		ArrayList<Event> criticalPath = calculateCriticalPath(events);

		ArrayList<ArrayList<Usage>> usages = makeListOfUsages();

		ArrayList<Worker> workers = new ArrayList<Worker>();
		for (int i = 0; i < numberOfWorkers; i++) {
			workers.add(new Worker(steals.get(i), leaps.get(i), spawns.get(i), usages.get(i)));
		}

		return new PlotPanel(startTime, endTime, workers, criticalPath);
	}

	private ArrayList<Event> calculateCriticalPath(ArrayList<Event> events) {
		Collections.sort(events);
		ArrayList<Event> criticalPath = new ArrayList<Event>();

		Event lastEvent = null;

		// The last '2' I guess is the starting point. -kff
		for (int i = events.size() - 1; i > -1; i--) {
			Event e = events.remove(i);
			if (e.typeOfEvent == 2) {
				criticalPath.add(e);
				lastEvent = e;
				break;
			}
		}

		for (int i = events.size() - 1; i > -1; i--) {
			Event event = events.get(i);
			if (lastEvent.typeOfEvent == 1 && event.worker == lastEvent.worker) {
				// The event before a '1' is the previous '1xy' in the same worker -kff
				if (event.typeOfEvent >= 100) {
					criticalPath.add(event);
					lastEvent = event;
				}
			} else if ((lastEvent.typeOfEvent == 2 || lastEvent.typeOfEvent == 5 || lastEvent.typeOfEvent == 6 || lastEvent.typeOfEvent == 3)
					&& event.worker == lastEvent.worker) {
				// From a '2', the previous event must be a '1' or a '4', I think. -kff
				// I think the same applies for both 5 and 6 as well -ph
				// And actually also for a 3, because if there is a 3 in the critical path, it comes from a leapfrog which didn't actually
				// involve any waiting. The next time the critical path moves should then involve a successful steal or leapfrogging -ph
				if (event.typeOfEvent == 1) {
					criticalPath.add(event);
					lastEvent = event;
				} else if (event.typeOfEvent == 4) {
					criticalPath.add(event);
					lastEvent = event;
				}
			} else if (lastEvent.typeOfEvent >= 100) {
				// it is either the reason that this worker started stealing (that would be a '3' if it is leap frogging and otherwise a '2' I guess)
				// *or* the spawn
				// ('5') of the stolen task (in which case the critical path moves to the victim), whatever happened latest.

				if (event.typeOfEvent == 3 && event.worker == lastEvent.worker) { // if it's leapfrogging

					// If it's it leapfrogging, it's not part of the critical path anymore, the next event should be a 5 or 6 in the worker
					// which stole the task which caused this to leapfrog.
					for (int j = i; j >= 0; j--) {
						event = events.get(j);
						if ((event.typeOfEvent == 5 || event.typeOfEvent == 6 || event.typeOfEvent == 1)
								&& event.worker == lastEvent.typeOfEvent - 100) { // this will catch the first spawn, sync or successful steal
							criticalPath.add(event);
							lastEvent = event;
						}
					}

					// If a worker is out stealing, it should not be in the critical path
					// } else if (event.typeOfEvent == 2 && event.worker == lastEvent.worker) { // I guess
					// criticalPath.add(event);
					// lastEvent = event;
				} else if (event.typeOfEvent == 5 && event.worker == lastEvent.typeOfEvent - 100) { // *or* the spawn of the stolen task
					criticalPath.add(event);
					lastEvent = event;
				}

			} else if (lastEvent.typeOfEvent == 4) {

				int thiefWhichCausedTheLeapfrog = -1;
				Event threeInTheThief = null;
				int indexForThreeInTheThief = -1;
				for (int j = i; j >= 0; j--) {
					Event tempEvent = events.get(j);
					if ((tempEvent.typeOfEvent >= 100 && lastEvent.worker == tempEvent.worker)) {
						thiefWhichCausedTheLeapfrog = tempEvent.typeOfEvent - 100;
					} else if (tempEvent.typeOfEvent == 3 && lastEvent.worker == tempEvent.worker) {
						threeInTheThief = tempEvent;
						indexForThreeInTheThief = j;
					}

					if (threeInTheThief != null && thiefWhichCausedTheLeapfrog != -1) {
						break;
					}
				}

				for (int j = i; j >= 0; j--) {
					Event tempEvent = events.get(j);
					if (tempEvent.typeOfEvent == 2 && tempEvent.worker == thiefWhichCausedTheLeapfrog) {
						// or in rare cases the '3' before the '4' in the victim, if that happened at about the same time that the thief signaled
						// completion.
						if (tempEvent.timeStamp.subtract(threeInTheThief.timeStamp).abs().compareTo(new BigDecimal(10000)) < 0) {
							criticalPath.add(threeInTheThief);
							lastEvent = threeInTheThief;
							i = indexForThreeInTheThief; // set the big for-loop to continue from where I found the 2
						} else {
							criticalPath.add(tempEvent);
							lastEvent = tempEvent;
							i = j; // set the big for loop to continue from where I found the 2
						}
						break;
					}
				}
			}
		}

		for (int i = 0; i < events.size(); i++) {
			if (events.get(i).typeOfEvent == 5) {
				criticalPath.add(events.get(i)); // the first spawn should be in there, just for good measure
				break;
			}
		}
		Collections.sort(criticalPath);// to leave it in chronological order
		return criticalPath;
	}

	private ArrayList<Event> makeListOfEvents() throws IOException {
		ArrayList<Event> events = new ArrayList<Event>();

		Event lastStealAttempt = null;

		String line;
		while ((line = bufferedReader.readLine()) != null) {
			if (line.matches("<.*>")) {
				return events;
			} else {
				if (line.matches("EVENT.*")) {
					line = line.replaceAll("[ ]+", ":");
					String[] tokens = line.split(":");
					int worker = Integer.parseInt(tokens[1]);
					int typeOfEvent = Integer.parseInt(tokens[2]);
					String timeStamp = tokens[3];

					if (typeOfEvent >= 100) {
						lastStealAttempt = new Event(new BigDecimal(timeStamp), worker, typeOfEvent);
					} else {
						if (lastStealAttempt != null) {
							events.add(lastStealAttempt);
							lastStealAttempt = null;
						}
						events.add(new Event(new BigDecimal(timeStamp), worker, typeOfEvent));
					}
				} else if (line.isEmpty()) { // Blank line
					continue;
				} else {
					return events;
				}
			}
		}
		return events;
	}

	private void readCrap() throws IOException {
		String line;
		while ((line = bufferedReader.readLine()) != null) {
			if (line.equalsIgnoreCase("---BEGIN STARTENDTIMES---")) {
				return;
			} else {
				System.out.println(line);
			}
		}
	}

	private ArrayList<ArrayList<Spawn>> makeListOfSpawns() throws NumberFormatException, IOException {
		ArrayList<ArrayList<Spawn>> spawns = new ArrayList<ArrayList<Spawn>>();
		for (int i = 0; i < numberOfWorkers; i++) {
			spawns.add(new ArrayList<Spawn>());
		}
		String line;
		while ((line = bufferedReader.readLine()) != null) {
			if (line.matches("<.*>")) {
				return spawns;
			} else if (!line.matches("[0-9]*:[0-9]*:.*")) {
				System.out.println(line);
			} else {
				String[] tokens = line.split(":");
				int spawner = Integer.parseInt(tokens[0]);
				BigDecimal timeStamp = new BigDecimal(tokens[1]);
				String taskName = tokens[2];
				spawns.get(spawner).add(new Spawn(spawner, timeStamp, taskName));
			}
		}
		return spawns;
	}

	private ArrayList<ArrayList<Usage>> makeListOfUsages() throws NumberFormatException, IOException {
		ArrayList<ArrayList<Usage>> usages = new ArrayList<ArrayList<Usage>>();
		for (int i = 0; i < numberOfWorkers; i++) {
			usages.add(new ArrayList<Usage>());
		}
		BigDecimal timeStampDiff = null;
		BigDecimal firstTimeStamp = null;
		String line;
		while ((line = bufferedReader.readLine()) != null) {
			String[] tokens = line.split(":");
			int worker = Integer.parseInt(tokens[0]);
			BigDecimal timeStamp = new BigDecimal(tokens[1]);
			double percentage = Double.parseDouble(tokens[2]);

			// This whole thing is to ensure that the first CPU usage is included
			if (timeStamp.compareTo(startTime) > 0) {
				usages.get(worker).add(new Usage(percentage, timeStamp));
			} else if (firstTimeStamp == null) {
				firstTimeStamp = timeStamp;
			} else if (timeStampDiff == null) {
				timeStampDiff = timeStamp.subtract(firstTimeStamp);
			} else {
				if (timeStamp.add(timeStampDiff).compareTo(startTime) > 0 && timeStamp.compareTo(endTime) < 0) {
					usages.get(worker).add(new Usage(percentage, timeStamp));
				}
			}
		}
		return usages;
	}

	private void setTheStartTimeToTheFirstSpawn(ArrayList<ArrayList<Spawn>> spawns) {
		BigDecimal firstSpawnTime = endTime;
		for (ArrayList<Spawn> spawnList : spawns) {
			if (!spawnList.isEmpty()) {
				firstSpawnTime = min(Collections.min(spawnList).timeStamp, firstSpawnTime);
			}
		}
		if (!firstSpawnTime.equals(endTime)) {
			startTime = startTime.add(firstSpawnTime.subtract(startTime));
		}
	}

	private BigDecimal min(BigDecimal first, BigDecimal second) {
		if (first.compareTo(second) <= 0) {
			return first;
		} else {
			return second;
		}
	}

	private ArrayList<ArrayList<Steal>> makeListOfSteals() throws NumberFormatException, IOException {

		ArrayList<ArrayList<Steal>> steals = new ArrayList<ArrayList<Steal>>();

		for (int i = 0; i < numberOfWorkers; i++) {
			steals.add(new ArrayList<Steal>());
		}

		String line;
		while ((line = bufferedReader.readLine()) != null) {
			if (line.matches("[0-9]*:[0-9]*:[0-9]*")) {
				String[] tokens = line.split(":");
				int thief = Integer.parseInt(tokens[0]);
				int victim = Integer.parseInt(tokens[1]);
				BigDecimal timeStamp = new BigDecimal(tokens[2]);
				steals.get(thief).add(new Steal(timeStamp, thief, victim));
			} else {
				return steals;
			}
		}
		return steals;
	}

}
